/* Toy32, a 32-bit miniset C compiler */
/* Code by Ren Olsthoorn. */
/* E-mail till June 1997: reneo@htsa.hva.nl */

#include <stdio.h>
#include <string.h>
#include <io.h>
#include <stdlib.h>
#include <ctype.h>

#define true -1
#define false 0

#define CR 13
#define LF 10
#define TAB 9

typedef unsigned char uchar;
typedef unsigned char byte;
typedef unsigned short word;
typedef unsigned long ulong;
typedef char str9[10];
typedef char str32[33];
typedef char str64[65];
typedef char string[255];

string program_name;    // programname without extention
FILE *fp_input;         // input filestream
FILE *fp_output;        // output filestream

char lookahead;         // lookahead char
string current_string;  // string loaded so far
long current_token;
long current_number;    // if token is number, then this is the number

string getted_name;
string getted_string_const;
string getted_newlabel;
string getted_srcln;

str64 current_procname; //name of current procedure if MAIN isn't declared yet

long linecount;  //how many C lines compiled so far
long labelcount; //generator for new label numbers

char hexcodestr[] = "0123456789ABCDEF";
char deccodestr[] = "0123456789";
char charcodestr[] = "_0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
char skipcodestr[] = "\r\n\t "; //CR+LF+TAB+SPACE

enum token { T_unknown,T_string_const,T_numeric_const,T_name,T_type,T_directive,
             T_while,T_if,T_else,T_do,T_for,
             T_begin,T_end,T_lparen,T_rparen,T_lbrack,T_rbrack,
	     T_separator,T_period,T_comma,
	     T_plus,T_min,T_mul,T_div,T_mod,T_pow,
	     T_assign,T_colon,
	     T_linecomment,T_comment,
	     T_equal,T_greater,T_less,T_less_equal,T_greater_equal,T_not_equal,
	     T_plusplus,T_minmin,
	     T_assembler,T_main,T_return,T_dataseg,
	     T_addsymbol,T_asm,T_include,
	     T_cout,T_streamout,T_streamin,T_endl,
	     T_lasttoken };

#define token_num T_lasttoken

str32 tokenname[token_num] = {
	"","","","","","",
	"WHILE","IF","ELSE","DO","FOR", //program flow tokens

	"{","}","(",")","[","]",        //compound tokens

	";",".",",",                    //separators

	"+","-","*","/","%","^",        //operators

	"=",":",

	"//","/*",                      //comment

	"==",">","<","<=",">=","!=",    //logical
	
	"++","--",

	"ASM","MAIN","RETURN","DATASEG",
	
	"ADDSYMBOL","","INCLUDE",

	"COUT","<<",">>","ENDL" };

//asmtable page 254 till 420 in 'Handboek voor 80386-programmeurs', KLUWER.
str9 asmtable[] = {
    "AAA", "AAD", "AAM", "AAS", "ADC", "ADD", "AND", "ARPL", "BOUND",
    "BSF", "BSR", "BT", "BTC", "BTR", "BTS",
    "CALL", "CBW", "CWDE", "CLC", "CLD", "CLI", "CLTS", "CMC", "CMP",
    "CMPS", "CMPSB", "CMPSW", "CMPSD", "CWD", "CDQ",
    "DAA", "DAS", "DEC", "DIV",
    "ENTER",
    "HLT",
    "IDIV", "IMUL", "IN", "INC", "INS", "INSB", "INSW", "INSD", "INT", "INTO",
    "IRET", "IRETD",
    "JA", "JAE", "JB", "JBE", "JC", "JCXZ", "JECXZ", "JE", "JZ", "JG", "JGE",
    "JL", "JLE", "JNA", "JNAE","JNB", "JNBE", "JNC", "JNE", "JNG", "JNGE",
    "JNL", "JNLE", "JNO", "JNP","JNS", "JNZ", "JO", "JP", "JPE", "JPO", "JS",
    "JZ", "JMP",
    "LAHF", "LAR", "LEA", "LEAVE", "LGDT", "LIDT", "LGS", "LSS", "LDS", "LES",
    "LFS", "LLDT", "LMSW", "LOCK", "LODS", "LODSB", "LODSW", "LODSD", "LOOP",
    "LOOPE", "LOOPZ", "LOOPNE", "LOOPNZ", "LSL", "LTR",
    "MOV", "MOVS", "MOVSB", "MOVSW", "MOVSD", "MOVSX", "MOVZX", "MUL",
    "NEG", "NOP", "NOT", "OR", "OUT", "OUTS", "OUTSB", "OUTSW", "OUTSD",
    "POP", "POPA", "POPAD", "POPF", "POPFD", "PUSH", "PUSHA", "PUSHAD",
    "PUSHF", "PUSHFD",
    "RCL", "RCR", "ROL", "ROR", "REP", "REPE", "REPZ", "REPNE", "REPNZ",
    "RET", "RETF", "RETN",
    "SAHF", "SAL", "SAR", "SHL", "SHR", "SBB", "SCAS", "SCASB", "SCASW",
    "SCASD", "SETA", "SETAE", "SETB", "SETBE", "SETC", "SETE", "SETG", "SETGE",
    "SETL", "SETLE", "SETNA", "SETNAE", "SETNB", "SETNBE", "SETNC", "SETNE", 
    "SETNG", "SETNGE", "SETNL", "SETNLE", "SETNO", "SETNP", "SETNS", "SETNZ",
    "SETO", "SETP", "SETPE", "SETPO", "SETS", "SETZ", "SGDT", "SIDT", "SHLD",
    "SHRD", "SLDT", "SMSW", "STC", "STD", "STI", "STOS", "STOSB",
    "STOSW", "STOSD", "STR", "SUB",
    "TEST",
    "VERR", "VERW",
    "WAIT",
    "XCHG", "XLAT", "XLATB", "XOR",
    "DB", "DW", "DD"
     };

#define asmtable_size sizeof(asmtable)/sizeof(str9)


enum pcode { P_call,P_return,P_clear,P_loadconst,P_loadvar,P_push,P_popadd,
             P_popsub,P_popmul,P_popdiv,P_store,P_putlabel,
	     P_jumpto,P_ifjumpto,P_ifnotjumpto,
	     P_programinit,P_programexit,
	     P_logical,P_logical_not,
	     P_greater,P_less,P_write_value,P_putcrlf,P_putstring,
	     P_loadoffset, P_loadpointer, P_storepointer,
	     P_bracket_offset, P_load_array_element, P_save_array_element,
	     P_incvar, P_decvar };


/*********************** Symbol Table *************************************/
struct inttype
{
	str32 name;
	word size;      //size of integer type in bytes
};

struct symbol
{
	str32 name;
	short kind;     //defines the type of symbol
	short elem1;    //general purpose element 1
	short elem2;    //general purpose element 2
	long  elem3;    //general purpose element 3
	unsigned char dump; //dump symbol or not
};

struct kindtable_element
{
        str32 kindname;
        long kindnr;
};

/*

KIND:  K_procedure
ELEM1: number of parameters
ELEM2: boolean, true = return parameter
ELEM3: -

KIND:  K_libproc
ELEM1: number of parameters
ELEM2: boolean, true = return parameter
ELEM3: boolean, true = extrn is declared, false = extrn is not declared

KIND:  K_pointer
ELEM1: -
ELEM2: size in bytes of the symbol where it's an pointer to
ELEM3: -

KIND:  K_stack
ELEM1: number of parameters
ELEM2: -
ELEM3: -

KIND:  K_array
ELEM1: number of elements
ELEM2: size in bytes of an array element
ELEM3: -

KIND:  K_long
ELEM1: -
ELEM2: -
ELEM3: -

*/

symbol found_symbol;
long found_symbol_index;

inttype typevoid    = { "_VOID", 0 };
inttype typechar    = { "_CHAR", 1 };
inttype typebyte    = { "_BYTE", 1 };
inttype typeshort   = { "_SHORT" , 2 };
inttype typeword    = { "_WORD" , 2 };
inttype typelong    = { "_LONG", 4 };

/* kind definitions */
#define K_void    0
#define K_char    1
#define K_byte    2
#define K_short   3
#define K_word    4
#define K_long    5

#define K_stack     100
#define K_procedure 101
#define K_libproc   102
#define K_array     103
#define K_pointer   104

//kindtable, for use with the #addsymbol directive
kindtable_element kindtable[] = { {"K_void", K_void},
                                  {"K_char", K_char},
                                  {"K_byte", K_byte},
                                  {"K_short", K_short},
                                  {"K_word", K_word},
                                  {"K_long", K_long},
                                  {"K_stack", K_stack},
                                  {"K_procedure", K_procedure},
                                  {"K_libproc", K_libproc},
                                  {"K_array", K_array},
                                  {"K_pointer", K_pointer} };
                         
#define kindtable_size sizeof(kindtable)/sizeof(kindtable_element)

symbol symboltable[512];
long symbolcount;

inttype typetable[512];
long typecount;

string stringconst[63];
long stringcount;
/**************************************************************************/

void printsymbols(void);

char* strcatc(char *s, char c) //add char c at the end of string s
{
        s[strlen(s)+1] = 0;
        s[strlen(s)] = c;
        return(s);
}

void quit(char *s)
{
        printsymbols();
        printf("\n%s\nExit on line %d.\n", s, linecount);
        exit(-1);
}

void expected(char *s)
{
        printsymbols();
        printf("\n%s expected on line %d.\n", s, linecount);
        exit(-1);
}

void readchar(void)
{
	if (!feof(fp_input))
	  lookahead = getc(fp_input);
	else
	  lookahead = ';';

	if (lookahead == CR)
	  linecount++;
}

void skipspace(void)
{
        char *tmp;
        
	while (((tmp = strchr(skipcodestr,lookahead)) != NULL) && (!feof(fp_input)))
	  readchar();
}

void getsrcln(void) //GetSourceLine is used by the ASM and comment statements
{
        char *tmp, tmpc;
        
        strcpy(getted_srcln,"");
        while ((lookahead != CR) && (lookahead != LF))
          {
          strcatc(getted_srcln, lookahead);
          readchar();
          }
        tmpc = getted_srcln[strlen(getted_srcln)-1]; //last character
	while (((tmp = strchr(skipcodestr,tmpc)) != NULL) && (strlen(getted_srcln) > 0))
	  {
          getted_srcln[strlen(getted_srcln)-1] = 0;
          tmpc = getted_srcln[strlen(getted_srcln)-1]; //last character
          }
        while ((lookahead == CR) || (lookahead == LF))
          readchar();
}

int findtype(char *name);
void linecommentproc(void);
void commentblock(void);

void readtoken(void)
{
	int i, j, kind;
	string tmpa;
	char *tmp;

	strcpy(current_string,"");
	current_token = T_unknown;

	current_number = 0;
	skipspace();

        //decimal?
	if ((tmp = strchr(deccodestr,lookahead)) != NULL)
	  {
	  while ((tmp = strchr(deccodestr,lookahead)) != NULL)
	    {
	    current_number = current_number * 10 + (tmp - deccodestr);
	    readchar();
	    }
	  current_token = T_numeric_const;
	  return;
	  }
	
	//name?
        if ((tmp = strchr(charcodestr,lookahead)) != NULL)
	  {
	  while ((tmp = strchr(charcodestr,lookahead)) != NULL)
	    {
	    strcatc(current_string, toupper(lookahead));
	    readchar();
	    }
	  //maybe it's an label
	  if (lookahead == ':')
	    {
	    strcatc(current_string, toupper(lookahead));
	    readchar();
	    current_token = T_asm;
	    return;
	    }
	  //maybe it's a normal token
	  for (i = 0; i < token_num; i++)
            if (strcmp(current_string, tokenname[i]) == 0)
              {
	      current_token = token(i);
	      return;
	      }
	  //maybe it's an assembly statement
	  for (i = 0; i < asmtable_size; i++)
	    if (strcmp(current_string, asmtable[i]) == 0)
	      {
	      current_token = T_asm;
	      return;
	      }
	  strcpy(tmpa, "_");
	  strcat(tmpa, current_string);

          //maybe it's an type, like long, byte, etc...
	  kind = findtype(tmpa);
	  if (kind != -1)
	    {
            current_token = T_type;
            return;
            }
	    
	  if (current_token == T_unknown)
	    current_token = T_name;
	  return;
	  }
	  
	  switch(lookahead)
	    {
	    case '(':
              readchar();
              current_token = T_lparen;
              break;
              
            case '#':
              readchar();
              current_token = T_directive;
              break;
              
            case '\'':
              //character value
              readchar();
              current_number = (long)lookahead;
              readchar();
              readchar();
	      current_token = T_numeric_const;
              break;

	    case '"':
              readchar();
              strcpy(current_string,"");
              i = false;

              do
                {
                switch (lookahead)
		  {
		  case CR:
		    quit("String exceeds line!");
		    break;

		  case '"':
		    readchar();
		    if (lookahead != '"')
	              i = true;
		    else
                      {
	              strcatc(current_string,lookahead);
                      readchar();
                      }
		    break;

		  default :
		    strcatc(current_string,lookahead);
		    readchar();
		    break;
		  }
		} while (i != true);
		
              current_token = T_string_const;
              break;

	    case '$': //HEXADECIMAL NUMBER
              readchar();

              while ((tmp = strchr(hexcodestr,toupper(lookahead))) != NULL)
		{
		current_number = (current_number << 4) + (tmp - hexcodestr);
		readchar();
		}
              current_token = T_numeric_const;
              break;

	    default:
              strcatc(current_string, toupper(lookahead));
              readchar();

              strcpy(tmpa, current_string);
              strcatc(tmpa, toupper(lookahead));

              /* Max. 2 char token */

              j = 0;
              for (i = 0; i < token_num; i++)
		if (strcmp(tmpa, tokenname[i]) == 0)
		  j = i;

              if (j != 0)
		readchar();
              else
		{
		j = 0;
		for (i = 0; i < token_num; i++)
		  if (strcmp(current_string, tokenname[i]) == 0)
		    j = i;
		}
              current_token = token(j);
              
              if (current_token == T_linecomment)
                linecommentproc();
              else
                if (current_token == T_comment)
                  commentblock();
              
              break;
	    } //switch
}

char* getname(void)
{
        if (current_token == T_name)
          {
          strcpy(getted_name,"_");
          strcat(getted_name, strupr(current_string));
	  }
	else
	  expected("Name");
	readtoken();
	return(getted_name);
}

long getnumber(void)
{
        long i;
        i = current_number;
        readtoken();
        return(i);
}

void addsymbol(char *name, int kind, short elem1 = 0, short elem2 = 0, long elem3 = 0, unsigned char dump = 1)
{
        strcpy(symboltable[symbolcount].name, name);  // name of symbol
        symboltable[symbolcount].kind   = kind;   // kind of symbol
        symboltable[symbolcount].elem1  = elem1;
        symboltable[symbolcount].elem2  = elem2;
        symboltable[symbolcount].elem3  = elem3;
        symboltable[symbolcount].dump = dump;
        symbolcount++;
}

int findsymbol(char *name)
{
        int i;

        for (i = 0; i < symbolcount; i++)
          if (strcmp(symboltable[i].name,name) == 0)
            {
            found_symbol = symboltable[i];
            found_symbol_index = i;
            return(symboltable[i].kind);
            }

        return(-1);
}

short findkind(char *kindname)
{
        long i,j = -1;
        for (i = 0; i < kindtable_size; i++)
          if (strcmp(kindname, strupr(kindtable[i].kindname)) == 0)
            j = kindtable[i].kindnr;
        return(j);
}

int lookmain(void)
{
        if (findsymbol("MAIN") == -1)
          return(false);
        else
          return(true);
}

void dumpsymbols(void)
{
        int i;

        fprintf(fp_output, "\n");
        fprintf(fp_output, ";---------------------- Variable Area --------------------------\n");

        for (i = 0; i < symbolcount; i++)
          switch(symboltable[i].kind)
            {
            case K_char:
            case K_byte:
            case K_short:
            case K_word:
            case K_long:
              if (symboltable[i].dump)
                fprintf(fp_output, "%s\tDB\t%d\tDUP (?)\n", symboltable[i].name, \
                        typetable[symboltable[i].kind].size);
              break;
            case K_stack:
              fprintf(fp_output, "%s\tEQU\tDWORD PTR [EBP+4+%d*4]\n", \
                                  symboltable[i].name, \
                                  symboltable[i].elem1);
              break;
            case K_array:
              if (symboltable[i].dump)
                fprintf(fp_output, "%s\tDB\t%d\tDUP (?)\n", symboltable[i].name, \
                        symboltable[i].elem1 * symboltable[i].elem2);
              break;
            case K_pointer:
              if (symboltable[i].dump)
                fprintf(fp_output, "%s\tDD\t1\t\n", symboltable[i].name);
              break;
            } //switch
}

void printsymbols(void)
{
        long i;
        
        printf("\nSymboltable: \n");
        for (i = 0; i < symbolcount; i++)
          {
          switch(symboltable[i].kind)
            {
            case K_void:
              printf("VOID    ");
              break;
            case K_procedure:
              printf("PROC    ");
              break;
            case K_libproc:
              printf("LIBPROC ");
              break;
            case K_char:
              printf("CHAR    ");
              break;
            case K_byte:
              printf("BYTE    ");
              break;
            case K_short:
              printf("SHORT   ");
              break;
            case K_word:
              printf("WORD    ");
              break;
            case K_long:
              printf("LONG    ");
              break;
            case K_stack:
              printf("STACK   ");
              break;
            case K_array:
              printf("ARRAY   ");
              break;
            case K_pointer:
              printf("POINTER ");
              break;
            } //switch
          printf("%s\n", symboltable[i].name);
          }
}

int findtype(char *name)
{
	long i;
	for (i = 0; i < typecount; i++)
	  if (strcmp(typetable[i].name, name) == 0)
	    return(i);
	    
	return(-1);
}

char* save_string_const(char *s)
{
	string tmp;
	strcpy(stringconst[stringcount], s);
	strcpy(getted_string_const, "_STR");
	ltoa (stringcount, tmp, 10);
	strcat(getted_string_const, tmp);
        stringcount++;
        return(getted_string_const);
}

void dumpstrings(void)
{
	int i;
	byte j;
	string s;
	string tmp, tmp2;

	fprintf(fp_output,"\n");
	fprintf(fp_output,";---------------------- String constants -----------------------\n");

	for (i = 0; i < stringcount; i++)
	  {
	  strcpy(s, stringconst[i]);
	  fprintf(fp_output,"_STR");
	  ltoa(i, tmp2, 10);
	  
	  fprintf(fp_output, tmp2);
	  strcpy(tmp2, "\tDB\t\'");
	  fprintf(fp_output, tmp2);

	  for (j = 0; j < strlen(s); j++)
	    if (s[j] != '\'')
              fprintf(fp_output, "%c", s[j]);
            else
              fprintf(fp_output, "\'\'");
          fprintf(fp_output,"\',0\n");
	  }

	fprintf(fp_output,"\n");
}

void emit(char *s)
{
	fprintf(fp_output,"\t%s", s);
}

void emitln(char *s)
{
	emit(s);
	fprintf(fp_output,"\n");
}

char* newlabel(void)
{
        string tmp;

        ltoa(labelcount, tmp, 10);
        labelcount++;
        
        strcpy(getted_newlabel, "L");
        strcat(getted_newlabel, tmp);
        return(getted_newlabel);
}

char* sizer(char *s, long size, char size_type)
{
        strcpy(s, "");
        
        switch(toupper(size_type))
          {
          case 'A':
          case 'B':
          case 'C':
          case 'D':
            switch(size)
              {
              case 1:
                strcatc(s,size_type);
                strcatc(s,'L');
                break;
              case 2:
                strcatc(s,size_type);
                strcatc(s,'X');
                break;
              case 4:
                strcatc(s,'E');
                strcatc(s,size_type);
                strcatc(s,'X');
                break;
              default:
                quit("Illegal size");
                break;
              }
            break;
          case 'P':
          case 'p':
            switch(size)
              {
              case 1:
                strcat(s,"BYTE PTR ");
                break;
              case 2:
                strcat(s,"WORD PTR ");
                break;
              case 4:
                strcat(s,"DWORD PTR ");
                break;
              default:
                quit("Illegal size");
                break;
              }
            break;
          }
          return(s);
}

int expression(void);

int emitcode(pcode c, long n = 0, char *s = "")
{
	string tmp, tmp2, tmp3;
	int x,y;

	switch(c)
	  {
	  case P_loadoffset:
	    fprintf(fp_output, "\tLEA   EAX, [%s]\n", s);
	    break;

	  case P_loadpointer:
	    fprintf(fp_output, "\tMOV   EAX, DWORD PTR [%s]\n", s);
	    break;

	  case P_storepointer:
	    fprintf(fp_output, "\tMOV   DWORD PTR [%s], EAX\n", s);
	    break;

          case P_bracket_offset:
            emitcode(P_loadconst, n); //n = size of one element in bytes
            emitcode(P_push);
            expression();
            emitcode(P_popmul);
            emitcode(P_push);
            break;
            
          case P_load_array_element:
            emitln("POP   EBX");  //bracket offset
            emitln("ADD   EBX,EAX");
            emitln("XOR   EAX,EAX");
            //      MOV   AL, BYTE PTR [EBX]
	    fprintf(fp_output, "\tMOV   %s,%s [EBX]\n", sizer(tmp,n,'A'), sizer(tmp2,n,'P'));
            break;

          case P_save_array_element:
            emitln("MOV   ECX,EAX"); //move data to ecx
            emitln("POP   EBX");  //var offset
            emitln("POP   EAX");  //bracket offset
            emitln("ADD   EBX,EAX");
            //      MOV   BYTE PTR [EBX],CL
	    fprintf(fp_output, "\tMOV   %s [EBX],%s\n", sizer(tmp,n,'P'), sizer(tmp2,n,'C'));
            break;

	  case P_call:
	    strcpy(tmp, "CALL  NEAR PTR ");
	    strcat(tmp, s);
	    emitln(tmp);
	    break;

	  case P_return:
	    fprintf(fp_output, "\tRETN  %d\n", n);
	    break;

	  case P_clear:
	    emitln("XOR   EAX,EAX");
	    break;

	  case P_loadconst:
	    fprintf(fp_output, "\tMOV   EAX, %d\n", n);
	    break;
	    
	  case P_decvar:
	  case P_incvar:
            if (c == P_incvar) strcpy(tmp,"INC");
            else strcpy(tmp, "DEC");
            switch(x)
              {
              case K_stack:
                fprintf(fp_output,"\t%s   [%s]\n",tmp,s);
                break;
              default:
                x = findsymbol(s);
                y = typetable[x].size;
                fprintf(fp_output,"\t%s   %s [%s]\n",tmp,sizer(tmp2,y,'P'),s);
                break;
              } //switch
	    break;
	    
          case P_loadvar:
            x = findsymbol(s);
            switch(x)
              {
              case K_stack:
                fprintf(fp_output,"\tMOV   EAX, %s\n",s);
                break;
              default:
                y = typetable[x].size;
                if (y < 4) emitln("XOR   EAX,EAX");
                fprintf(fp_output,"\tMOV   %s, %s [%s]\n",sizer(tmp,y,'A'),sizer(tmp2,y,'P'),s);
                break;
              } //switch
            break;
            
          case P_push:
            emitln("PUSH  EAX");
            break;
            
          case P_popadd:
            emitln("POP   EBX");
            emitln("ADD   EAX,EBX");
            break;
            
	  case P_popsub:
            emitln("POP   EBX");
            emitln("SUB   EBX,EAX");
            emitln("MOV   EAX,EBX");
            break;
                    
          case P_popmul:
            emitln("POP   EBX");
            emitln("MUL   EBX");
            break;
            
          case P_popdiv:
            emitln("MOV   EBX,EAX");
            emitln("XOR   EDX,EDX");
            emitln("POP   EAX");
            emitln("DIV   EBX");
            break;
            
          case P_store:
            x = findsymbol(s);
            switch(x)
              {
              case K_stack:
                fprintf(fp_output,"\tMOV   %s, EAX\n",s);
                break;
              default:
	        y = typetable[x].size;
	        fprintf(fp_output,"\tMOV   %s [%s], %s\n",sizer(tmp,y,'P'),s,sizer(tmp2,y,'A'));
	        break;
	      }
	    break;

	  case P_putlabel:
            fprintf(fp_output,"%s:\n",s);
	    break;

	  case P_jumpto:
            fprintf(fp_output,"\tJMP   %s\n",s);
	    break;

	  case P_ifjumpto:
	    emitln("TEST  EAX,EAX");
            fprintf(fp_output,"\tJNZ   %s\n",s);
	    break;

	  case P_ifnotjumpto:
	    emitln("TEST  EAX,EAX");
            fprintf(fp_output,"\tJZ    %s\n",s);
	    break;


          case P_programinit:
            fprintf(fp_output,";%s.ASM is generated by Toy32.\n", s);
            fprintf(fp_output,";Thanks to Thomas Pytel for PMODE.       \n");
            fprintf(fp_output,"                                         \n");
            fprintf(fp_output,".386p                                    \n");
            fprintf(fp_output,"locals                                   \n");
            fprintf(fp_output,"                                         \n");
            fprintf(fp_output,"include pmc.inc                          \n");
            fprintf(fp_output,"public  _PMmain                          \n");
            fprintf(fp_output,"                                         \n");
            fprintf(fp_output,"_TEXTseg                                 \n");
            fprintf(fp_output,"                                         \n");
            fprintf(fp_output,";----------------------  Code start ----------------------------\n");
            fprintf(fp_output,"_PMmain:                                 \n");
            fprintf(fp_output,"      STI                                \n");
            fprintf(fp_output,"      JMP MAIN                           \n");
            break;

          case P_programexit:
            emitln("RET");
            fprintf(fp_output, "ends\n");
            break;

          case P_logical:
            emitln("NEG   EAX");         // AX <> 0 ---> Carry
            emitln("MOV   EAX,-1");      //       0 ---> AX
            emitln("ADC   EAX,0");       // Carry  ----> ALL AX
            break;

          case P_logical_not:
            emitln("NEG   EAX");         // AX <> 0 ---> Carry
            emitln("MOV   EAX,-1");      //      -1 ---> AX
            emitln("ADC   EAX,1");       // Not Carry -> ALL AX
            break;

          case P_greater:
            emitln("POP   EBX");
            emitln("SUB   EAX,EBX");
            emitln("MOV   EAX,0" );
            emitln("SBB   EAX,0" );
            break;

          case P_less:
            emitln("POP   EBX");
            emitln("SUB   EBX,EAX");
            emitln("MOV   EAX,0" );
            emitln("SBB   EAX,0" );
            break;

          case P_write_value:
            emitln("CALL  L_WriteEAX");
            break;

          case P_putcrlf:
            emitln("CALL  L_DoCR");
            break;

          case P_putstring:
            fprintf(fp_output,"\tLEA   EDX, [%s]\n",s);
            emitln("CALL  L_WriteStr");
            break;
            
          default:
            quit("Unknown pcode");
            break;
	  }
	return(0);
}

short iscompareop(long x)
{
        if ((x >= T_equal) && (x <= T_not_equal))
          return(true);
        else
          return(false);
}

short isaddop(long x)
{
        if ((x >= T_plus) && (x <= T_min))
          return(true);
        else
          return(false);
}

short ismulop(long x)
{
        if ((x >= T_mul) && (x <= T_div))
          return(true);
        else
          return(false);
}

void match(long x)
{
        if (current_token != x)
          {
          if (x <= token_num)
            expected(tokenname[x]);
          else
            quit("unknown token expected, compiler error!");
          }
        else
          readtoken();
}

void matchstop(long x)
{
        if (current_token != x)
          {
          if (x <= token_num)
	    expected(tokenname[x]);
	  else
	    quit("unknown token expected, compiler error!");
	  }
}

void skip(long x)
{
        while (current_token != x)
          readtoken();
        match(x);
}

void skipstop(long x)
{
        while (current_token != x)
          readtoken();
}

void check_local(char *s)
{
        str64 name;

	if (lookmain() == false)
	  {
	  //no main!
	  strcpy(name, current_procname);
	  strcatc(name, '@');
	  strcat(name, s);
	  if (findsymbol(name) != -1)
	    strcpy(s, name);
	  }
}

int procedure_code(short inputkind, char *name)
{
        str64 name2, name3;
        long old_index;
        
        strcpy(name2, name);
        if ((inputkind == K_procedure) || (inputkind == K_libproc))
          {
          old_index = found_symbol_index;
          
          match(T_lparen);
          while (current_token != T_rparen)
            {
            expression();
            emitcode(P_push); //push parameter
            if (current_token == T_comma)
              readtoken();
            }
          match(T_rparen);
          
          if (inputkind == K_libproc)
            {
            strcpy(name3, "L");
            strcat(name3, name2);
            strcpy(name2, name3);
            
            //look if extrn is already defined
            if (symboltable[old_index].elem3 == false)
              {
              //extrn not yet defined
              fprintf(fp_output, "extrn %s:proc\n", name2);
              symboltable[old_index].elem3 = true;
              }
            }
            
          emitcode(P_call,0,name2);
          return(true);
          }
        else
          return(false);
}

int value(void)
{
	int kind = -1;
	str64 name;

        switch(current_token)
          {
          case T_lparen:
	    match(T_lparen);
	    kind = expression();
	    match(T_rparen);
            break;
            
          case T_string_const:
            save_string_const(current_string);
            emitcode(P_loadoffset, 0, getted_string_const);
            match(T_string_const);
            break;
          
          case T_name:
	    strcpy(name, getname());
            kind = findsymbol(name);
            if (!procedure_code(kind, name))
              {
              check_local(name);
              kind = findsymbol(name);
              if (kind == K_array || kind == K_pointer)
                {
                if (current_token == T_lbrack)
                  {
                  //array element
                  match(T_lbrack);
                  findsymbol(name);
                  emitcode(P_bracket_offset, found_symbol.elem2);
                  switch(kind)
                    {
                    case K_array:
                      emitcode(P_loadoffset, 0, name);
                      break;
                    case K_pointer:
                      emitcode(P_loadpointer, 0, name);
                      break;
                    }
                  findsymbol(name);
                  emitcode(P_load_array_element, found_symbol.elem2);
                  match(T_rbrack);
                  }
                else
                  switch(kind)
                    {
                    case K_array:
                      emitcode(P_loadoffset, 0, name);
                      break;
                    case K_pointer:
                      emitcode(P_loadpointer, 0, name);
                      break;
                    }
                } //if (kind == ARRAY OR POINTER)
              else
	        kind = emitcode(P_loadvar,0,name);
	      }
            break;
            
          case T_numeric_const:
	    kind = emitcode(P_loadconst,getnumber());
            break;
            
          default:
	    quit("error in expression");
	    break;
          }
          
	return(kind);
}

void factor(void)
{
	long tmp;
	int kind = value();
        
        while (iscompareop(current_token))
          {
          emitcode(P_push);
          tmp = current_token;
          match(tmp);
          value();

          switch(tmp)
            {
            case T_equal:
              emitcode(P_popsub,     kind,"");
              emitcode(P_logical,    kind,"");
              break;
              
            case T_not_equal:
              emitcode(P_popsub,     kind,"");
              emitcode(P_logical_not,kind,"");
              break;
              
            case T_greater:
              emitcode(P_greater,     kind,"");
              break;
              
            case T_less:
              emitcode(P_less,        kind,"");
              break;
              
            case T_greater_equal:
              emitcode(P_less,        kind,"");
              emitcode(P_logical_not,kind,"");
              break;
              
            case T_less_equal:
              emitcode(P_greater,     kind,"");
              emitcode(P_logical_not,kind,"");
              break;
            } //switch
          } //while
}

void multiply(void)
{
        match(T_mul);
        factor();
        emitcode(P_popmul,0,"");
}

void divide(void)
{
        match(T_div);
        factor();
	emitcode(P_popdiv,0,"");
}

void term(void)
{
        factor();
	while (ismulop(current_token))
          {
          emitcode(P_push,0,"");
          switch(current_token)
            {
            case T_mul:
              multiply();
              break;
            case T_div:
              divide();
              break;
            }
          }
}

void add(void)
{
        match(T_plus);
        term();
        emitcode(P_popadd,0,"");
}

void subtract(void)
{
        match(T_min);
        term();
        emitcode(P_popsub,0,"");
}

int expression(void)
{
        int kind = -1;

        if (isaddop(current_token))
          emitcode(P_clear,0,"");
        else
          term();

        while (isaddop(current_token))
          {
          emitcode(P_push,0,"");
          switch(current_token)
            {
            case T_plus:
              add();
              break;
            case T_min:
              subtract();
              break;
            } //switch
          } //while
          
        return(kind);
}

void statement(void);

void assignment(void)
{
        string name_var;
        strcpy(name_var, getname());
        short kind;
        
        kind = findsymbol(name_var);
        
        if (!procedure_code(kind, name_var))
          {
          check_local(name_var);
          switch(current_token)
            {
            case T_plusplus:
              match(T_plusplus);
              emitcode(P_incvar,0,name_var);
              break;
              
            case T_minmin:
              match(T_minmin);
              emitcode(P_decvar,0,name_var);
              break;
              
            case T_lbrack:
              //array storage
              match(T_lbrack);
              findsymbol(name_var);
              emitcode(P_bracket_offset, found_symbol.elem2);
              findsymbol(name_var);
              switch(found_symbol.kind)
                {
                case K_array:
                  emitcode(P_loadoffset, 0, name_var);
                  break;
                case K_pointer:
                  emitcode(P_loadpointer, 0, name_var);
                  break;
                }
              emitcode(P_push);
              match(T_rbrack);
              match(T_assign);
              expression();
              findsymbol(name_var);
              emitcode(P_save_array_element, found_symbol.elem2);
              break;
              
            default:
              match(T_assign);
              expression();
              findsymbol(name_var);
              if (found_symbol.kind == K_pointer)
                emitcode(P_storepointer,0,name_var);
              else
                emitcode(P_store,0,name_var);
              break;
            } //switch
          } //if
}

void commentblock(void)
{
        short stop;
        do
          {
          while (lookahead != '*')
            readchar();
          readchar();
          if (lookahead == '/')
            stop = true;
          } while (stop != true);
        skipspace();
        readchar();
        readtoken();
}

void linecommentproc(void)
{
	getsrcln();
        readtoken();
}

void asm_block(void)
{
        match(T_assembler);
        matchstop(T_begin);
	getsrcln();
	while (strchr(getted_srcln,'}') == 0)
	  {
	  fprintf(fp_output,"%s\n", getted_srcln);
	  getsrcln();
	  }
        readtoken();
}

void while_loop(void)
{
        str32 testlabel, donelabel;
        
        match(T_while);

        strcpy(testlabel, newlabel());
        strcpy(donelabel, newlabel());

        emitcode(P_putlabel,0,testlabel);
        expression();
        emitcode(P_ifnotjumpto,0,donelabel);

        statement();
        emitcode(P_jumpto,0,testlabel);

        emitcode(P_putlabel,0,donelabel);
}

void c_for_loop(void)
{
        str32 donelabel, statementlabel, counterlabel, testlabel;
        string index, limit;

        match(T_for);                           // for
        match(T_lparen);                        // (
  
        assignment();                          // f = 0
        match(T_separator);                     // ;
  
        strcpy(testlabel, newlabel());
        strcpy(donelabel, newlabel());
        strcpy(statementlabel, newlabel());
        strcpy(counterlabel, newlabel());
  
        //expression
        emitcode(P_putlabel,0,testlabel);         
        expression();                          // f < 10
        match(T_separator);                     // ;
        emitcode(P_ifnotjumpto,0,donelabel);
        emitcode(P_jumpto,0,statementlabel);

        //increment
        emitcode(P_putlabel,0,counterlabel);
        statement();                           // f = f + 1
        match(T_rparen);                        // )
        emitcode(P_jumpto,0,testlabel);

        //statement
        emitcode(P_putlabel,0,statementlabel);
        statement();                           //loop routine
        emitcode(P_jumpto,0,counterlabel);

        emitcode(P_putlabel,0,donelabel);
}

void if_then_else(void)
{
        str32 elselabel, donelabel;

        match(T_if);
        strcpy(elselabel, newlabel());
        strcpy(donelabel, newlabel());
        expression();
        emitcode(P_ifnotjumpto,0,elselabel);
        statement();
        if (current_token != T_else)
          emitcode(P_putlabel,0,elselabel);
        else
          {
          match(T_else);
          emitcode(P_jumpto,0,donelabel);
          emitcode(P_putlabel,0,elselabel);
          statement();
          }
        emitcode(P_putlabel,0,donelabel);
}

void blockstatement(void)
{
        match(T_begin);
        while (current_token != T_end)
          {
          if (current_token == T_separator)
            readtoken();
          else
            statement();
          }
        match(T_end);
}

void do_while_loop(void)
{
        str32 start;

        match(T_do);
        strcpy(start, newlabel());
        emitcode(P_putlabel,0,start);
        statement();
        match(T_while);
        expression();
        emitcode(P_ifjumpto,0,start);
        match(T_separator);
}

void cout_work(void)
{
        string sx;

        match(T_cout);
        if (current_token == T_streamout) //no operands
          {
          match(T_streamout);
          do
            {
            if (current_token == T_endl)
              {
              match(T_endl);
              emitcode(P_putcrlf,0,"");
              }
            else
              {
              if (current_token == T_string_const)
                {
                strcpy(sx, save_string_const(current_string));
                match(T_string_const);
                emitcode(P_putstring,0,sx);
                }
              else
                {
                expression();
                emitcode(P_write_value,0,"");
                }
              }
            if (current_token != T_separator)
              match(T_streamout);
            } while (current_token != T_separator);
          } //if
        match(T_separator);
}

void return_statement(void)
{
        short kind;
        
        match(T_return);
        if (lookmain() == true)
          quit("main cannot yet have an return type!");
        kind = findsymbol(current_procname);
        if (found_symbol.elem2)
          expression();
        //exit procedure
        findsymbol(current_procname);
        emitln("POP   EBP");
        emitcode(P_return,found_symbol.elem1*4,"");
        match(T_separator);
}

void dataseg_statement(void)
{
        match(T_dataseg);
        matchstop(T_begin);
        fprintf(fp_output, "_DATAseg\n");
	getsrcln();
	while (strchr(getted_srcln,'}') == 0)
	  {
	  fprintf(fp_output,"%s\n", getted_srcln);
	  getsrcln();
	  }
        fprintf(fp_output, "ends\n\n");
        readtoken();
}

void type_block(short status);

void statement(void)
{
        str64 name;
        short kind;

        switch(current_token)
          {
          case T_while:
            while_loop();
            break;
          
          case T_do:
            do_while_loop();
            break;
            
          case T_for:
            c_for_loop();
            break;
            
          case T_if:
            if_then_else();
            break;
            
          case T_begin:
            blockstatement();
            break;
            
          case T_assembler:
            asm_block();
            break;
                
          case T_cout:
            cout_work();
            break;
            
          case T_return:
            return_statement();
            break;
            
          case T_type:
            type_block(1);
	    break;

          case T_dataseg:
            dataseg_statement();
            break;
	  
	  case T_separator:
	    break;
	  
	  case T_asm:
	    emit(current_string);
	    getsrcln();
	    emitln(getted_srcln);
	    readtoken();
	    break;
	    
	  default:
            assignment();
            break;
          }
}

void storeparam(char *s, short &paramnr)
{
        str64 tmp, tmp2;

        if (current_token == T_name)
          {   
          strcpy(tmp, getname());

          if (current_token == T_comma)
            {
            match(T_comma);
            match(T_type);
            storeparam(s, paramnr);
            }
          
          strcpy(tmp2, s);
          strcatc(tmp2, '@'); //parameter signature
          strcat(tmp2, tmp);
          addsymbol(tmp2,K_stack,paramnr);
          paramnr++;
          }
        else
          expected("NAME");
}

void type_block(short status)
//status = 0, define global variables
//status = 1, define local variables
{
        str64 name, name2;
        string procname;
        short kind, kind2;
        short paramnr;
        short size;
       
        strcpy(name,"_");
        strcat(name, strupr(current_string));
        kind = findtype(name);
        readtoken();
            
        while(current_token != T_separator)
          {
          if (current_token == T_main)
            return;

          if (current_token == T_mul)
            {
            //pointer definition
            match(T_mul);
            strcpy(procname, getname());
            if ((status == 1) && (lookmain() == false))
              {
              strcpy(name2, current_procname);
              strcatc(name2, '@');
              strcat(name2, procname);
              strcpy(procname, name2);
              }
            addsymbol(procname, K_pointer, 0, typetable[kind].size);
            continue;
            }
          
          if (current_token == T_name)
            {
            strcpy(procname, getname());

            switch(current_token)
              {
              case T_lparen:
                //procedure definitie
                match(T_lparen);
                strcpy(current_procname, procname); //set current_procname
                paramnr = 1;
                switch(current_token)
                  {
                  case T_type:
                    strcpy(name2,"_");
                    strcat(name2, strupr(current_string));
                    match(T_type);
                    kind2 = findtype(name2);
                    if (kind2 != K_void) //skip if void parameter
                      {
                      if (current_token == T_name)
                        storeparam(procname,paramnr);
                      else
                        expected("NAME");
                      }
                    break;
                  }
                match(T_rparen);
                emitcode(P_putlabel,0,procname);
                emitln("PUSH  EBP");
                emitln("MOV   EBP,ESP");
                if (kind != K_void)
                  addsymbol(procname,K_procedure,paramnr-1,true);
                else
                  addsymbol(procname,K_procedure,paramnr-1,false);
                blockstatement();
                emitln("POP   EBP");
                emitcode(P_return,(paramnr-1)*4,"");
                return;
            
              case T_lbrack:
                //array definition
                strcpy(name, procname);
                match(T_lbrack);
                if (current_token != T_numeric_const)
                  expected("simple array-size number");
                size = current_number;
                addsymbol(name,K_array,size,typetable[kind].size);
                readtoken();
                match(T_rbrack);
                break;
          
              case T_comma:
              case T_separator:
                //var definition
                if ((status == 1) && (lookmain() == false))
                  {
                  strcpy(name2, current_procname);
                  strcatc(name2, '@');
                  strcat(name2, procname);
                  strcpy(procname, name2);
                  }
                addsymbol(procname,kind);
                break;
              
              default:
                readtoken();
                break;
              } //switch
            } //if
          else
            if (current_token != T_separator)
              readtoken();

          } //while
}

void directive_statement(void);

void compile_code_before_MAIN(void)
{
        short done = false;

        while (done != true)
          {
          switch(current_token)
            {
	    case T_separator:
              match(T_separator);
              break;
              
            case T_type:        //procedure or variables
              type_block(0);    //global typeblock
              break;     
              
            case T_directive:
              directive_statement();
              break;
              
            case T_dataseg:
              dataseg_statement();
              break;

            case T_assembler:
              asm_block();
              break;
              
            default:
	      done = true;
              break;
            } //switch
          }
}

void compile_includefile(char *filename)
{
        FILE *old_fp_input;
        long old_linecount;
        char old_lookahead;
        
        old_fp_input = fp_input;
        old_linecount = linecount;
        old_lookahead = lookahead;

	if ((fp_input = fopen(filename, "rb")) == NULL)
	  quit("Can't open include file!");

        readchar();
        readtoken();
        compile_code_before_MAIN();

        fclose(fp_input);

        lookahead = old_lookahead;
        linecount = old_linecount;        
        fp_input = old_fp_input;
}

long get_immediate_value(void)
{
        if (current_token == T_min)
          {
          readtoken();
          return(-current_number);
          }
        else
          return(current_number);
}

void directive_statement(void)
{
        string tmps,kinds;
        long kind, elem1, elem2, elem3, dump;

        match(T_directive);
        switch(current_token)
          {
          case T_addsymbol:
            match(T_addsymbol);
            match(T_lparen);
            strcpy(tmps, current_string);
            match(T_string_const);
            match(T_comma);
            kind = findkind(current_string);
            if (kind == -1)
              quit("unknown kind!");
            match(T_name);
            match(T_comma);
            elem1 = get_immediate_value();
            match(T_numeric_const);
            match(T_comma);
            elem2 = get_immediate_value();
            match(T_numeric_const);
            match(T_comma);
            elem3 = get_immediate_value();
            match(T_numeric_const);
            match(T_comma);
            dump = get_immediate_value();
            match(T_numeric_const);
            match(T_rparen);
            match(T_separator);
            
            addsymbol(tmps,kind,elem1,elem2,elem3,dump);
            break;
          case T_include:
            match(T_include);
            if (current_token != T_string_const)
              expected("include-filename between quotes");
            compile_includefile(current_string);
            readtoken(); //read next token after "INCLUDE FILENAME"
            break;
            
          default:
            quit("Unknown directive!");
            break;
          }
}

void compile(void)
{
        emitcode(P_programinit,0,program_name);
        compile_code_before_MAIN();
        skip(T_main);
        match(T_lparen);
        skip(T_rparen);
        emitcode(P_putlabel,0,"MAIN");
        addsymbol("MAIN",K_void);
        blockstatement();
        emitcode(P_programexit,0,"");
        
        //any data?
        if (current_token == T_dataseg)
          dataseg_statement();

        //symbols
        fprintf(fp_output, "_DATAseg\n");
        dumpsymbols();
        dumpstrings();
        fprintf(fp_output, "ends\n");

        fprintf(fp_output, "end\n");
}

void init(void)
{
	labelcount = 0;
	symbolcount = 0;
	stringcount = 0;
	linecount = 0;

	typetable[0] = typevoid;
	typetable[1] = typechar;
	typetable[2] = typebyte;
	typetable[3] = typeshort;
	typetable[4] = typeword;
	typetable[5] = typelong;
        typecount = 6;
        
        readchar();
        readtoken();
}

int main(int argc, char *argv[])
{
	char *tmp;
	string output_name;

        printf("Toy32 V2.0, a one pass, 32-bit, subset C compiler.\n");

	if (argc < 2)
	  {
	  printf("\nUsage : toy32 filename\n");
	  return(-1);
	  }

	if ((fp_input = fopen(argv[1], "rb")) == NULL)
	  {
	  printf("Can't open file %s!\n",argv[1]);
	  return(-1);
	  }

        //chop filename from point
        if ((tmp = strchr(argv[1],'.')) != NULL)
	  tmp[0] = 0;
	  
	strcpy(program_name,argv[1]);
	strcpy(output_name, argv[1]);
	strcat(output_name,".ASM");

	if ((fp_output = fopen(output_name, "wt")) == NULL)
	  {
	  printf("Can't open file %s!\n",output_name);
	  return(-1);
	  }
	  
	init();
	compile();
        fclose(fp_input);

        printf("%d lines compiled succesfully...\n", linecount);
	return(0);
}
